/*
Copyright (C) 2001 Lkb

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include <ddraw.h>
#include <d3dx.h>
#include "resource.h"

#include "DBGConsole.h"

#include "SurfaceHandler.h"
#include "GfxContext.h"

#include <map>

static map<D3DCOLOR, SurfaceHandler*> singleColorTexturesMap;

static DWORD maxSingleColorTextures = 0;

template<typename T>
static void DrawSingleColorTexture(SurfaceHandler* pSurface, T color)
{
	DrawInfo di;
	pSurface->StartUpdate(&di);

	if((di.dwWidth == 1) && (di.dwHeight == 1))
	{
		*(T*)di.lpSurface = color;
	}
	else
	{
		void* maxlpSurfaceHeight = (void*)((UINT_PTR)di.lpSurface + di.dwHeight * di.lPitch);
		UINT yIncrement = di.lPitch - (di.dwWidth * sizeof(T));
		for(; di.lpSurface < maxlpSurfaceHeight; di.lpSurface = (void*)((UINT_PTR)di.lpSurface + yIncrement))
		{
			void* maxlpSurfaceWidth = (void*)((UINT_PTR)di.lpSurface + di.dwWidth * sizeof(T));
			for(; di.lpSurface < maxlpSurfaceWidth; di.lpSurface = (void*)((UINT_PTR)di.lpSurface + sizeof(T)))
			{
				*(T*)di.lpSurface = color;
			}
		}
	}

	pSurface->EndUpdate(&di);
}

static void ReduceSingleColorTextureMemoryUsage()
{
	if(singleColorTexturesMap.size() > maxSingleColorTextures)
	{
		DBGConsole_Msg(0, "Single-color textures map cleared!");
		// TODO: use a better strategy
		for(map<D3DCOLOR, SurfaceHandler*>::iterator iter = singleColorTexturesMap.begin(); iter != singleColorTexturesMap.end(); iter++)
		{
			delete iter->second;
		}
		singleColorTexturesMap.clear();
	}
}

LPDIRECTDRAWSURFACE7 GetSingleColorTexture(D3DCOLOR color)
{
	ReduceSingleColorTextureMemoryUsage();

	SurfaceHandler*& pSurface = singleColorTexturesMap[color];
	if(pSurface != NULL)
		return pSurface->GetSurface();

	//DBGConsole_Msg(0, "Single-color texture created: 0x%08X!", color);

	DDSCAPS2 ddsCaps2;
	ZeroMemory(&ddsCaps2, sizeof(ddsCaps2));
	ddsCaps2.dwCaps = DDSCAPS_TEXTURE; 
	
	if(maxSingleColorTextures == 0)
	{
		// TODO: is locking really required?
		DWORD totalMem;
		DWORD freeMemBefore;
		g_GfxContext.Lock();
		g_GfxContext.GetDD()->GetAvailableVidMem(&ddsCaps2, &totalMem, &freeMemBefore);
		g_GfxContext.Unlock();

		pSurface = new SurfaceHandler(1, 1);
		
		DWORD freeMemAfter;
		g_GfxContext.Lock();
		g_GfxContext.GetDD()->GetAvailableVidMem(&ddsCaps2, NULL, &freeMemAfter);
		g_GfxContext.Unlock();

		// must be at least 256 so if the driver reports allocation incorrectly, video memory is less likely to be exhausted
		maxSingleColorTextures = max(totalMem / (max(freeMemBefore - freeMemAfter, (DWORD)256) * 64), (DWORD)16);

		//DBGConsole_Msg(0, "Single-color texture memory usage:  %u bytes", freeMemBefore - freeMemAfter);
		//DBGConsole_Msg(0, "Max number of single-color texture: %u bytes", maxSingleColorTextures);
	}
	else
	{
		pSurface = new SurfaceHandler(1, 1);
	}

	switch(pSurface->GetPixelSize())
	{
		case 4:
			DrawSingleColorTexture<DWORD>(pSurface, (DWORD)color);
			break;
		case 2:
			// 0xAaBbCcDd ->
			// 0x0000ABCD
			DrawSingleColorTexture<WORD>(pSurface,
				(WORD)(
					((color >> (4 * 1)) & 0xf) |	// A
					((color >> (4 * 2)) & 0xf0) |	// B
					((color >> (4 * 3)) & 0xf00) |	// C
					((color >> (4 * 4)) & 0xf000) |	// D
					0
				)
				);
			break;
	}

	return pSurface->GetSurface();
}

void DropSingleColorTextures()
{
	for(map<D3DCOLOR, SurfaceHandler*>::iterator iter = singleColorTexturesMap.begin(); iter != singleColorTexturesMap.end(); iter++)
	{
		delete iter->second;
	}
	singleColorTexturesMap.clear();	
}
